@@ -1,68 +1,92 @@
 
 #line 1 "parser.rl"
-#include "ruby.h"
-#include "unicode.h"
-#if HAVE_RE_H
-#include "re.h"
-#endif
-#if HAVE_RUBY_ST_H
-#include "ruby/st.h"
-#endif
-#if HAVE_ST_H
-#include "st.h"
-#endif
+#include "parser.h"
 
-#define EVIL 0x666
+/* unicode */
 
-#ifndef RHASH_TBL
-#define RHASH_TBL(hsh) (RHASH(hsh)->tbl)
-#endif
+static const char digit_values[256] = { 
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, -1,
+    -1, -1, -1, -1, -1, -1, 10, 11, 12, 13, 14, 15, -1, -1, -1, -1, -1, -1, -1,
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+    10, 11, 12, 13, 14, 15, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+    -1, -1, -1, -1, -1, -1, -1
+};
+
+static UTF32 unescape_unicode(const unsigned char *p)
+{
+    char b;
+    UTF32 result = 0;
+    b = digit_values[p[0]];
+    if (b < 0) return UNI_REPLACEMENT_CHAR;
+    result = (result << 4) | b;
+    b = digit_values[p[1]];
+    result = (result << 4) | b;
+    if (b < 0) return UNI_REPLACEMENT_CHAR;
+    b = digit_values[p[2]];
+    result = (result << 4) | b;
+    if (b < 0) return UNI_REPLACEMENT_CHAR;
+    b = digit_values[p[3]];
+    result = (result << 4) | b;
+    if (b < 0) return UNI_REPLACEMENT_CHAR;
+    return result;
+}
+
+static int convert_UTF32_to_UTF8(char *buf, UTF32 ch) 
+{
+    int len = 1;
+    if (ch <= 0x7F) {
+        buf[0] = (char) ch;
+    } else if (ch <= 0x07FF) {
+        buf[0] = (char) ((ch >> 6) | 0xC0);
+        buf[1] = (char) ((ch & 0x3F) | 0x80);
+        len++;
+    } else if (ch <= 0xFFFF) {
+        buf[0] = (char) ((ch >> 12) | 0xE0);
+        buf[1] = (char) (((ch >> 6) & 0x3F) | 0x80);
+        buf[2] = (char) ((ch & 0x3F) | 0x80);
+        len += 2;
+    } else if (ch <= 0x1fffff) {
+        buf[0] =(char) ((ch >> 18) | 0xF0);
+        buf[1] =(char) (((ch >> 12) & 0x3F) | 0x80);
+        buf[2] =(char) (((ch >> 6) & 0x3F) | 0x80);
+        buf[3] =(char) ((ch & 0x3F) | 0x80);
+        len += 3;
+    } else {
+        buf[0] = '?';
+    }
+    return len;
+}
 
 #ifdef HAVE_RUBY_ENCODING_H
-#include "ruby/encoding.h"
-#define FORCE_UTF8(obj) rb_enc_associate((obj), rb_utf8_encoding())
+static VALUE CEncoding_ASCII_8BIT, CEncoding_UTF_8, CEncoding_UTF_16BE,
+    CEncoding_UTF_16LE, CEncoding_UTF_32BE, CEncoding_UTF_32LE;
+static ID i_encoding, i_encode, i_encode_bang, i_force_encoding;
 #else
-#define FORCE_UTF8(obj)
+static ID i_iconv;
 #endif
 
 static VALUE mJSON, mExt, cParser, eParserError, eNestingError;
 static VALUE CNaN, CInfinity, CMinusInfinity;
 
 static ID i_json_creatable_p, i_json_create, i_create_id, i_create_additions,
-          i_chr, i_max_nesting, i_allow_nan, i_object_class, i_array_class; 
-
-#define MinusInfinity "-Infinity"
-
-typedef struct JSON_ParserStruct {
-    VALUE Vsource;
-    char *source;
-    long len;
-    char *memo;
-    VALUE create_id;
-    int max_nesting;
-    int current_nesting;
-    int allow_nan;
-    VALUE object_class;
-    VALUE array_class;
-} JSON_Parser;
-
-static char *JSON_parse_object(JSON_Parser *json, char *p, char *pe, VALUE *result);
-static char *JSON_parse_array(JSON_Parser *json, char *p, char *pe, VALUE *result);
-static char *JSON_parse_value(JSON_Parser *json, char *p, char *pe, VALUE *result);
-static char *JSON_parse_string(JSON_Parser *json, char *p, char *pe, VALUE *result);
-static char *JSON_parse_integer(JSON_Parser *json, char *p, char *pe, VALUE *result);
-static char *JSON_parse_float(JSON_Parser *json, char *p, char *pe, VALUE *result);
-
-#define GET_STRUCT                          \
-    JSON_Parser *json;                      \
-    Data_Get_Struct(self, JSON_Parser, json);
+          i_chr, i_max_nesting, i_allow_nan, i_symbolize_names, i_object_class,
+          i_array_class, i_key_p, i_deep_const_get;
 
 
-#line 84 "parser.rl"
+#line 108 "parser.rl"
 
 
 
-#line 66 "parser.c"
+#line 90 "parser.c"
 static const int JSON_object_start = 1;
 static const int JSON_object_first_final = 27;
 static const int JSON_object_error = 0;
@@ -70,7 +94,7 @@ static const int JSON_object_error = 0;
 static const int JSON_object_en_main = 1;
 
 
-#line 117 "parser.rl"
+#line 143 "parser.rl"
 
 
 static char *JSON_parse_object(JSON_Parser *json, char *p, char *pe, VALUE *result)
@@ -80,20 +104,20 @@ static char *JSON_parse_object(JSON_Parser *json, char *p, char *pe, VALUE *resu
     VALUE object_class = json->object_class;
 
     if (json->max_nesting && json->current_nesting > json->max_nesting) {
-        rb_raise(eNestingError, "nesting of %d is to deep", json->current_nesting);
+        rb_raise(eNestingError, "nesting of %d is too deep", json->current_nesting);
     }
 
     *result = NIL_P(object_class) ? rb_hash_new() : rb_class_new_instance(0, 0, object_class);
 
     
-#line 90 "parser.c"
+#line 114 "parser.c"
 	{
 	cs = JSON_object_start;
 	}
 
-#line 132 "parser.rl"
+#line 158 "parser.rl"
     
-#line 97 "parser.c"
+#line 121 "parser.c"
 	{
 	if ( p == pe )
 		goto _test_eof;
@@ -121,9 +145,12 @@ case 2:
 		goto st2;
 	goto st0;
 tr2:
-#line 103 "parser.rl"
+#line 127 "parser.rl"
 	{
-        char *np = JSON_parse_string(json, p, pe, &last_name);
+        char *np;
+        json->parsing_name = 1;
+        np = JSON_parse_string(json, p, pe, &last_name);
+        json->parsing_name = 0;
         if (np == NULL) { p--; {p++; cs = 3; goto _out;} } else {p = (( np))-1;}
     }
 	goto st3;
@@ -131,7 +158,7 @@ st3:
 	if ( ++p == pe )
 		goto _test_eof3;
 case 3:
-#line 135 "parser.c"
+#line 161 "parser.c"
 	switch( (*p) ) {
 		case 13: goto st3;
 		case 32: goto st3;
@@ -198,7 +225,7 @@ case 8:
 		goto st8;
 	goto st0;
 tr11:
-#line 92 "parser.rl"
+#line 116 "parser.rl"
 	{
         VALUE v = Qnil;
         char *np = JSON_parse_value(json, p, pe, &v); 
@@ -214,7 +241,7 @@ st9:
 	if ( ++p == pe )
 		goto _test_eof9;
 case 9:
-#line 218 "parser.c"
+#line 244 "parser.c"
 	switch( (*p) ) {
 		case 13: goto st9;
 		case 32: goto st9;
@@ -303,14 +330,14 @@ case 18:
 		goto st9;
 	goto st18;
 tr4:
-#line 108 "parser.rl"
+#line 134 "parser.rl"
 	{ p--; {p++; cs = 27; goto _out;} }
 	goto st27;
 st27:
 	if ( ++p == pe )
 		goto _test_eof27;
 case 27:
-#line 314 "parser.c"
+#line 340 "parser.c"
 	goto st0;
 st19:
 	if ( ++p == pe )
@@ -408,13 +435,13 @@ case 26:
 	_out: {}
 	}
 
-#line 133 "parser.rl"
+#line 159 "parser.rl"
 
     if (cs >= JSON_object_first_final) {
         if (RTEST(json->create_id)) {
             VALUE klassname = rb_hash_aref(*result, json->create_id);
             if (!NIL_P(klassname)) {
-                VALUE klass = rb_path2class(StringValueCStr(klassname));
+                VALUE klass = rb_funcall(mJSON, i_deep_const_get, 1, klassname);
                 if RTEST(rb_funcall(klass, i_json_creatable_p, 0)) {
                     *result = rb_funcall(klass, i_json_create, 1, *result);
                 }
@@ -427,7 +454,7 @@ case 26:
 }
 
 
-#line 431 "parser.c"
+#line 457 "parser.c"
 static const int JSON_value_start = 1;
 static const int JSON_value_first_final = 21;
 static const int JSON_value_error = 0;
@@ -435,7 +462,7 @@ static const int JSON_value_error = 0;
 static const int JSON_value_en_main = 1;
 
 
-#line 231 "parser.rl"
+#line 257 "parser.rl"
 
 
 static char *JSON_parse_value(JSON_Parser *json, char *p, char *pe, VALUE *result)
@@ -443,14 +470,14 @@ static char *JSON_parse_value(JSON_Parser *json, char *p, char *pe, VALUE *resul
     int cs = EVIL;
 
     
-#line 447 "parser.c"
+#line 473 "parser.c"
 	{
 	cs = JSON_value_start;
 	}
 
-#line 238 "parser.rl"
+#line 264 "parser.rl"
     
-#line 454 "parser.c"
+#line 480 "parser.c"
 	{
 	if ( p == pe )
 		goto _test_eof;
@@ -475,14 +502,14 @@ st0:
 cs = 0;
 	goto _out;
 tr0:
-#line 179 "parser.rl"
+#line 205 "parser.rl"
 	{
         char *np = JSON_parse_string(json, p, pe, result);
         if (np == NULL) { p--; {p++; cs = 21; goto _out;} } else {p = (( np))-1;}
     }
 	goto st21;
 tr2:
-#line 184 "parser.rl"
+#line 210 "parser.rl"
 	{
         char *np;
         if(pe > p + 9 && !strncmp(MinusInfinity, p, 9)) {
@@ -502,7 +529,7 @@ tr2:
     }
 	goto st21;
 tr5:
-#line 202 "parser.rl"
+#line 228 "parser.rl"
 	{ 
         char *np;
         json->current_nesting++;
@@ -512,7 +539,7 @@ tr5:
     }
 	goto st21;
 tr9:
-#line 210 "parser.rl"
+#line 236 "parser.rl"
 	{ 
         char *np;
         json->current_nesting++;
@@ -522,7 +549,7 @@ tr9:
     }
 	goto st21;
 tr16:
-#line 172 "parser.rl"
+#line 198 "parser.rl"
 	{
         if (json->allow_nan) {
             *result = CInfinity;
@@ -532,7 +559,7 @@ tr16:
     }
 	goto st21;
 tr18:
-#line 165 "parser.rl"
+#line 191 "parser.rl"
 	{
         if (json->allow_nan) {
             *result = CNaN;
@@ -542,19 +569,19 @@ tr18:
     }
 	goto st21;
 tr22:
-#line 159 "parser.rl"
+#line 185 "parser.rl"
 	{
         *result = Qfalse;
     }
 	goto st21;
 tr25:
-#line 156 "parser.rl"
+#line 182 "parser.rl"
 	{
         *result = Qnil;
     }
 	goto st21;
 tr28:
-#line 162 "parser.rl"
+#line 188 "parser.rl"
 	{
         *result = Qtrue;
     }
@@ -563,9 +590,9 @@ st21:
 	if ( ++p == pe )
 		goto _test_eof21;
 case 21:
-#line 218 "parser.rl"
+#line 244 "parser.rl"
 	{ p--; {p++; cs = 21; goto _out;} }
-#line 569 "parser.c"
+#line 595 "parser.c"
 	goto st0;
 st2:
 	if ( ++p == pe )
@@ -726,7 +753,7 @@ case 20:
 	_out: {}
 	}
 
-#line 239 "parser.rl"
+#line 265 "parser.rl"
 
     if (cs >= JSON_value_first_final) {
         return p;
@@ -736,7 +763,7 @@ case 20:
 }
 
 
-#line 740 "parser.c"
+#line 766 "parser.c"
 static const int JSON_integer_start = 1;
 static const int JSON_integer_first_final = 5;
 static const int JSON_integer_error = 0;
@@ -744,7 +771,7 @@ static const int JSON_integer_error = 0;
 static const int JSON_integer_en_main = 1;
 
 
-#line 255 "parser.rl"
+#line 281 "parser.rl"
 
 
 static char *JSON_parse_integer(JSON_Parser *json, char *p, char *pe, VALUE *result)
@@ -752,15 +779,15 @@ static char *JSON_parse_integer(JSON_Parser *json, char *p, char *pe, VALUE *res
     int cs = EVIL;
 
     
-#line 756 "parser.c"
+#line 782 "parser.c"
 	{
 	cs = JSON_integer_start;
 	}
 
-#line 262 "parser.rl"
+#line 288 "parser.rl"
     json->memo = p;
     
-#line 764 "parser.c"
+#line 790 "parser.c"
 	{
 	if ( p == pe )
 		goto _test_eof;
@@ -794,14 +821,14 @@ case 3:
 		goto st0;
 	goto tr4;
 tr4:
-#line 252 "parser.rl"
+#line 278 "parser.rl"
 	{ p--; {p++; cs = 5; goto _out;} }
 	goto st5;
 st5:
 	if ( ++p == pe )
 		goto _test_eof5;
 case 5:
-#line 805 "parser.c"
+#line 831 "parser.c"
 	goto st0;
 st4:
 	if ( ++p == pe )
@@ -820,7 +847,7 @@ case 4:
 	_out: {}
 	}
 
-#line 264 "parser.rl"
+#line 290 "parser.rl"
 
     if (cs >= JSON_integer_first_final) {
         long len = p - json->memo;
@@ -832,7 +859,7 @@ case 4:
 }
 
 
-#line 836 "parser.c"
+#line 862 "parser.c"
 static const int JSON_float_start = 1;
 static const int JSON_float_first_final = 10;
 static const int JSON_float_error = 0;
@@ -840,7 +867,7 @@ static const int JSON_float_error = 0;
 static const int JSON_float_en_main = 1;
 
 
-#line 286 "parser.rl"
+#line 312 "parser.rl"
 
 
 static char *JSON_parse_float(JSON_Parser *json, char *p, char *pe, VALUE *result)
@@ -848,15 +875,15 @@ static char *JSON_parse_float(JSON_Parser *json, char *p, char *pe, VALUE *resul
     int cs = EVIL;
 
     
-#line 852 "parser.c"
+#line 878 "parser.c"
 	{
 	cs = JSON_float_start;
 	}
 
-#line 293 "parser.rl"
+#line 319 "parser.rl"
     json->memo = p;
     
-#line 860 "parser.c"
+#line 886 "parser.c"
 	{
 	if ( p == pe )
 		goto _test_eof;
@@ -914,14 +941,14 @@ case 5:
 		goto st0;
 	goto tr7;
 tr7:
-#line 280 "parser.rl"
+#line 306 "parser.rl"
 	{ p--; {p++; cs = 10; goto _out;} }
 	goto st10;
 st10:
 	if ( ++p == pe )
 		goto _test_eof10;
 case 10:
-#line 925 "parser.c"
+#line 951 "parser.c"
 	goto st0;
 st6:
 	if ( ++p == pe )
@@ -982,7 +1009,7 @@ case 9:
 	_out: {}
 	}
 
-#line 295 "parser.rl"
+#line 321 "parser.rl"
 
     if (cs >= JSON_float_first_final) {
         long len = p - json->memo;
@@ -995,7 +1022,7 @@ case 9:
 
 
 
-#line 999 "parser.c"
+#line 1025 "parser.c"
 static const int JSON_array_start = 1;
 static const int JSON_array_first_final = 17;
 static const int JSON_array_error = 0;
@@ -1003,7 +1030,7 @@ static const int JSON_array_error = 0;
 static const int JSON_array_en_main = 1;
 
 
-#line 331 "parser.rl"
+#line 357 "parser.rl"
 
 
 static char *JSON_parse_array(JSON_Parser *json, char *p, char *pe, VALUE *result)
@@ -1012,19 +1039,19 @@ static char *JSON_parse_array(JSON_Parser *json, char *p, char *pe, VALUE *resul
     VALUE array_class = json->array_class;
 
     if (json->max_nesting && json->current_nesting > json->max_nesting) {
-        rb_raise(eNestingError, "nesting of %d is to deep", json->current_nesting);
+        rb_raise(eNestingError, "nesting of %d is too deep", json->current_nesting);
     }
     *result = NIL_P(array_class) ? rb_ary_new() : rb_class_new_instance(0, 0, array_class);
 
     
-#line 1021 "parser.c"
+#line 1047 "parser.c"
 	{
 	cs = JSON_array_start;
 	}
 
-#line 344 "parser.rl"
+#line 370 "parser.rl"
     
-#line 1028 "parser.c"
+#line 1054 "parser.c"
 	{
 	if ( p == pe )
 		goto _test_eof;
@@ -1063,7 +1090,7 @@ case 2:
 		goto st2;
 	goto st0;
 tr2:
-#line 312 "parser.rl"
+#line 338 "parser.rl"
 	{
         VALUE v = Qnil;
         char *np = JSON_parse_value(json, p, pe, &v); 
@@ -1079,7 +1106,7 @@ st3:
 	if ( ++p == pe )
 		goto _test_eof3;
 case 3:
-#line 1083 "parser.c"
+#line 1109 "parser.c"
 	switch( (*p) ) {
 		case 13: goto st3;
 		case 32: goto st3;
@@ -1179,14 +1206,14 @@ case 12:
 		goto st3;
 	goto st12;
 tr4:
-#line 323 "parser.rl"
+#line 349 "parser.rl"
 	{ p--; {p++; cs = 17; goto _out;} }
 	goto st17;
 st17:
 	if ( ++p == pe )
 		goto _test_eof17;
 case 17:
-#line 1190 "parser.c"
+#line 1216 "parser.c"
 	goto st0;
 st13:
 	if ( ++p == pe )
@@ -1242,73 +1269,88 @@ case 16:
 	_out: {}
 	}
 
-#line 345 "parser.rl"
+#line 371 "parser.rl"
 
     if(cs >= JSON_array_first_final) {
         return p + 1;
     } else {
         rb_raise(eParserError, "%u: unexpected token at '%s'", __LINE__, p);
+        return NULL;
     }
 }
 
-static VALUE json_string_unescape(char *p, char *pe)
+static VALUE json_string_unescape(VALUE result, char *string, char *stringEnd)
 {
-    VALUE result = rb_str_buf_new(pe - p + 1);
+    char *p = string, *pe = string, *unescape;
+    int unescape_len;
 
-    while (p < pe) {
-        if (*p == '\\') {
-            p++;
-            if (p >= pe) return Qnil; /* raise an exception later, \ at end */
-            switch (*p) {
-                case '"':
-                case '\\':
-                    rb_str_buf_cat(result, p, 1);
-                    p++;
-                    break;
-                case 'b':
-                    rb_str_buf_cat2(result, "\b");
-                    p++;
-                    break;
-                case 'f':
-                    rb_str_buf_cat2(result, "\f");
-                    p++;
-                    break;
+    while (pe < stringEnd) {
+        if (*pe == '\\') {
+            unescape = (char *) "?";
+            unescape_len = 1;
+            if (pe > p) rb_str_buf_cat(result, p, pe - p);
+            switch (*++pe) {
                 case 'n':
-                    rb_str_buf_cat2(result, "\n");
-                    p++;
+                    unescape = (char *) "\n";
                     break;
                 case 'r':
-                    rb_str_buf_cat2(result, "\r");
-                    p++;
+                    unescape = (char *) "\r";
                     break;
                 case 't':
-                    rb_str_buf_cat2(result, "\t");
-                    p++;
+                    unescape = (char *) "\t";
+                    break;
+                case '"':
+                    unescape = (char *) "\"";
+                    break;
+                case '\\':
+                    unescape = (char *) "\\";
+                    break;
+                case 'b':
+                    unescape = (char *) "\b";
+                    break;
+                case 'f':
+                    unescape = (char *) "\f";
                     break;
                 case 'u':
-                    if (p > pe - 4) { 
+                    if (pe > stringEnd - 4) { 
                         return Qnil;
                     } else {
-                        p = JSON_convert_UTF16_to_UTF8(result, p, pe, strictConversion);
+                        char buf[4];
+                        UTF32 ch = unescape_unicode((unsigned char *) ++pe);
+                        pe += 3;
+                        if (UNI_SUR_HIGH_START == (ch & 0xFC00)) {
+                            pe++;
+                            if (pe > stringEnd - 6) return Qnil;
+                            if (pe[0] == '\\' && pe[1] == 'u') {
+                                UTF32 sur = unescape_unicode((unsigned char *) pe + 2);
+                                ch = (((ch & 0x3F) << 10) | ((((ch >> 6) & 0xF) + 1) << 16)
+                                        | (sur & 0x3FF));
+                                pe += 5;
+                            } else {
+                                unescape = (char *) "?";
+                                break;
+                            }
+                        }
+                        unescape_len = convert_UTF32_to_UTF8(buf, ch);
+                        unescape = buf;
                     }
                     break;
                 default:
-                    rb_str_buf_cat(result, p, 1);
-                    p++;
-                    break;
+                    p = pe;
+                    continue;
             }
+            rb_str_buf_cat(result, unescape, unescape_len);
+            p = ++pe;
         } else {
-            char *q = p;
-            while (*q != '\\' && q < pe) q++;
-            rb_str_buf_cat(result, p, q - p);
-            p = q;
+            pe++;
         }
     }
+    rb_str_buf_cat(result, p, pe - p);
     return result;
 }
 
 
-#line 1312 "parser.c"
+#line 1353 "parser.c"
 static const int JSON_string_start = 1;
 static const int JSON_string_first_final = 8;
 static const int JSON_string_error = 0;
@@ -1316,24 +1358,24 @@ static const int JSON_string_error = 0;
 static const int JSON_string_en_main = 1;
 
 
-#line 429 "parser.rl"
+#line 470 "parser.rl"
 
 
 static char *JSON_parse_string(JSON_Parser *json, char *p, char *pe, VALUE *result)
 {
     int cs = EVIL;
 
-    *result = rb_str_new("", 0);
+    *result = rb_str_buf_new(0);
     
-#line 1329 "parser.c"
+#line 1370 "parser.c"
 	{
 	cs = JSON_string_start;
 	}
 
-#line 437 "parser.rl"
+#line 478 "parser.rl"
     json->memo = p;
     
-#line 1337 "parser.c"
+#line 1378 "parser.c"
 	{
 	if ( p == pe )
 		goto _test_eof;
@@ -1358,9 +1400,9 @@ case 2:
 		goto st0;
 	goto st2;
 tr2:
-#line 415 "parser.rl"
+#line 456 "parser.rl"
 	{
-        *result = json_string_unescape(json->memo + 1, p);
+        *result = json_string_unescape(*result, json->memo + 1, p);
         if (NIL_P(*result)) {
 			p--;
 			{p++; cs = 8; goto _out;}
@@ -1369,14 +1411,14 @@ tr2:
 			{p = (( p + 1))-1;}
 		}
 	}
-#line 426 "parser.rl"
+#line 467 "parser.rl"
 	{ p--; {p++; cs = 8; goto _out;} }
 	goto st8;
 st8:
 	if ( ++p == pe )
 		goto _test_eof8;
 case 8:
-#line 1380 "parser.c"
+#line 1421 "parser.c"
 	goto st0;
 st3:
 	if ( ++p == pe )
@@ -1452,8 +1494,11 @@ case 7:
 	_out: {}
 	}
 
-#line 439 "parser.rl"
+#line 480 "parser.rl"
 
+    if (json->symbolize_names && json->parsing_name) {
+      *result = rb_str_intern(*result);
+    }
     if (cs >= JSON_string_first_final) {
         return p + 1;
     } else {
@@ -1463,7 +1508,7 @@ case 7:
 
 
 
-#line 1467 "parser.c"
+#line 1511 "parser.c"
 static const int JSON_start = 1;
 static const int JSON_first_final = 10;
 static const int JSON_error = 0;
@@ -1471,7 +1516,7 @@ static const int JSON_error = 0;
 static const int JSON_en_main = 1;
 
 
-#line 473 "parser.rl"
+#line 517 "parser.rl"
 
 
 /* 
@@ -1486,6 +1531,54 @@ static const int JSON_en_main = 1;
  *
  */
 
+static VALUE convert_encoding(VALUE source)
+{
+    char *ptr = RSTRING_PTR(source);
+    long len = RSTRING_LEN(source);
+    if (len < 2) {
+        rb_raise(eParserError, "A JSON text must at least contain two octets!");
+    }
+#ifdef HAVE_RUBY_ENCODING_H
+    {
+        VALUE encoding = rb_funcall(source, i_encoding, 0);
+        if (encoding == CEncoding_ASCII_8BIT) {
+            if (len >= 4 &&  ptr[0] == 0 && ptr[1] == 0 && ptr[2] == 0) {
+                source = rb_str_dup(source);
+                rb_funcall(source, i_force_encoding, 1, CEncoding_UTF_32BE);
+                source = rb_funcall(source, i_encode_bang, 1, CEncoding_UTF_8);
+            } else if (len >= 4 && ptr[0] == 0 && ptr[2] == 0) {
+                source = rb_str_dup(source);
+                rb_funcall(source, i_force_encoding, 1, CEncoding_UTF_16BE);
+                source = rb_funcall(source, i_encode_bang, 1, CEncoding_UTF_8);
+            } else if (len >= 4 && ptr[1] == 0 && ptr[2] == 0 && ptr[3] == 0) {
+                source = rb_str_dup(source);
+                rb_funcall(source, i_force_encoding, 1, CEncoding_UTF_32LE);
+                source = rb_funcall(source, i_encode_bang, 1, CEncoding_UTF_8);
+            } else if (len >= 4 && ptr[1] == 0 && ptr[3] == 0) {
+                source = rb_str_dup(source);
+                rb_funcall(source, i_force_encoding, 1, CEncoding_UTF_16LE);
+                source = rb_funcall(source, i_encode_bang, 1, CEncoding_UTF_8);
+            } else {
+                FORCE_UTF8(source);
+            }
+        } else {
+            source = rb_funcall(source, i_encode, 1, CEncoding_UTF_8);
+        }
+    }
+#else
+    if (len >= 4 &&  ptr[0] == 0 && ptr[1] == 0 && ptr[2] == 0) {
+      source = rb_funcall(mJSON, i_iconv, 3, rb_str_new2("utf-8"), rb_str_new2("utf-32be"), source);
+    } else if (len >= 4 && ptr[0] == 0 && ptr[2] == 0) {
+      source = rb_funcall(mJSON, i_iconv, 3, rb_str_new2("utf-8"), rb_str_new2("utf-16be"), source);
+    } else if (len >= 4 && ptr[1] == 0 && ptr[2] == 0 && ptr[3] == 0) {
+      source = rb_funcall(mJSON, i_iconv, 3, rb_str_new2("utf-8"), rb_str_new2("utf-32le"), source);
+    } else if (len >= 4 && ptr[1] == 0 && ptr[3] == 0) {
+      source = rb_funcall(mJSON, i_iconv, 3, rb_str_new2("utf-8"), rb_str_new2("utf-16le"), source);
+    }
+#endif
+    return source;
+}
+
 /*
  * call-seq: new(source, opts => {})
  *
@@ -1503,6 +1596,9 @@ static const int JSON_en_main = 1;
  * * *allow_nan*: If set to true, allow NaN, Infinity and -Infinity in
  *   defiance of RFC 4627 to be parsed by the Parser. This option defaults to
  *   false.
+ * * *symbolize_names*: If set to true, returns symbols for the names
+ *   (keys) in a JSON object. Otherwise strings are returned, which is also
+ *   the default.
  * * *create_additions*: If set to false, the Parser doesn't create
  *   additions even if a matchin class and create_id was found. This option
  *   defaults to true.
@@ -1514,21 +1610,18 @@ static VALUE cParser_initialize(int argc, VALUE *argv, VALUE self)
     char *ptr;
     long len;
     VALUE source, opts;
-    GET_STRUCT;
+    GET_PARSER;
     rb_scan_args(argc, argv, "11", &source, &opts);
-    source = StringValue(source);
+    source = convert_encoding(StringValue(source));
     ptr = RSTRING_PTR(source);
     len = RSTRING_LEN(source);
-    if (len < 2) {
-        rb_raise(eParserError, "A JSON text must at least contain two octets!");
-    }
     if (!NIL_P(opts)) {
         opts = rb_convert_type(opts, T_HASH, "Hash", "to_hash");
         if (NIL_P(opts)) {
             rb_raise(rb_eArgError, "opts needs to be like a hash");
         } else {
             VALUE tmp = ID2SYM(i_max_nesting);
-            if (st_lookup(RHASH_TBL(opts), tmp, 0)) {
+            if (option_given_p(opts, tmp)) {
                 VALUE max_nesting = rb_hash_aref(opts, tmp);
                 if (RTEST(max_nesting)) {
                     Check_Type(max_nesting, T_FIXNUM);
@@ -1540,14 +1633,21 @@ static VALUE cParser_initialize(int argc, VALUE *argv, VALUE self)
                 json->max_nesting = 19;
             }
             tmp = ID2SYM(i_allow_nan);
-            if (st_lookup(RHASH_TBL(opts), tmp, 0)) {
+            if (option_given_p(opts, tmp)) {
                 VALUE allow_nan = rb_hash_aref(opts, tmp);
                 json->allow_nan = RTEST(allow_nan) ? 1 : 0;
             } else {
                 json->allow_nan = 0;
             }
+            tmp = ID2SYM(i_symbolize_names);
+            if (option_given_p(opts, tmp)) {
+                VALUE symbolize_names = rb_hash_aref(opts, tmp);
+                json->symbolize_names = RTEST(symbolize_names) ? 1 : 0;
+            } else {
+                json->symbolize_names = 0;
+            }
             tmp = ID2SYM(i_create_additions);
-            if (st_lookup(RHASH_TBL(opts), tmp, 0)) {
+            if (option_given_p(opts, tmp)) {
                 VALUE create_additions = rb_hash_aref(opts, tmp);
                 if (RTEST(create_additions)) {
                     json->create_id = rb_funcall(mJSON, i_create_id, 0);
@@ -1558,13 +1658,13 @@ static VALUE cParser_initialize(int argc, VALUE *argv, VALUE self)
                 json->create_id = rb_funcall(mJSON, i_create_id, 0);
             }
             tmp = ID2SYM(i_object_class);
-            if (st_lookup(RHASH_TBL(opts), tmp, 0)) {
+            if (option_given_p(opts, tmp)) {
                 json->object_class = rb_hash_aref(opts, tmp);
             } else {
                 json->object_class = Qnil;
             }
             tmp = ID2SYM(i_array_class);
-            if (st_lookup(RHASH_TBL(opts), tmp, 0)) {
+            if (option_given_p(opts, tmp)) {
                 json->array_class = rb_hash_aref(opts, tmp);
             } else {
                 json->array_class = Qnil;
@@ -1578,18 +1678,6 @@ static VALUE cParser_initialize(int argc, VALUE *argv, VALUE self)
         json->array_class = Qnil;
     }
     json->current_nesting = 0;
-    /*
-       Convert these?
-    if (len >= 4 &&  ptr[0] == 0 && ptr[1] == 0 && ptr[2] == 0) {
-        rb_raise(eParserError, "Only UTF8 octet streams are supported atm!");
-    } else if (len >= 4 && ptr[0] == 0 && ptr[2] == 0) {
-        rb_raise(eParserError, "Only UTF8 octet streams are supported atm!");
-    } else if (len >= 4 && ptr[1] == 0 && ptr[2] == 0 && ptr[3] == 0) {
-        rb_raise(eParserError, "Only UTF8 octet streams are supported atm!");
-    } else if (len >= 4 && ptr[1] == 0 && ptr[3] == 0) {
-        rb_raise(eParserError, "Only UTF8 octet streams are supported atm!");
-    }
-    */
     json->len = len;
     json->source = ptr;
     json->Vsource = source;
@@ -1607,19 +1695,19 @@ static VALUE cParser_parse(VALUE self)
     char *p, *pe;
     int cs = EVIL;
     VALUE result = Qnil;
-    GET_STRUCT;
+    GET_PARSER;
 
     
-#line 1614 "parser.c"
+#line 1701 "parser.c"
 	{
 	cs = JSON_start;
 	}
 
-#line 611 "parser.rl"
+#line 698 "parser.rl"
     p = json->source;
     pe = p + json->len;
     
-#line 1623 "parser.c"
+#line 1710 "parser.c"
 	{
 	if ( p == pe )
 		goto _test_eof;
@@ -1675,7 +1763,7 @@ case 5:
 		goto st1;
 	goto st5;
 tr3:
-#line 462 "parser.rl"
+#line 506 "parser.rl"
 	{
         char *np;
         json->current_nesting = 1;
@@ -1684,7 +1772,7 @@ tr3:
     }
 	goto st10;
 tr4:
-#line 455 "parser.rl"
+#line 499 "parser.rl"
 	{
         char *np;
         json->current_nesting = 1;
@@ -1696,7 +1784,7 @@ st10:
 	if ( ++p == pe )
 		goto _test_eof10;
 case 10:
-#line 1700 "parser.c"
+#line 1787 "parser.c"
 	switch( (*p) ) {
 		case 13: goto st10;
 		case 32: goto st10;
@@ -1753,16 +1841,17 @@ case 9:
 	_out: {}
 	}
 
-#line 614 "parser.rl"
+#line 701 "parser.rl"
 
     if (cs >= JSON_first_final && p == pe) {
         return result;
     } else {
         rb_raise(eParserError, "%u: unexpected token at '%s'", __LINE__, p);
+        return Qnil;
     }
 }
 
-inline static JSON_Parser *JSON_allocate()
+static JSON_Parser *JSON_allocate()
 {
     JSON_Parser *json = ALLOC(JSON_Parser);
     MEMZERO(json, JSON_Parser, 1);
@@ -1796,7 +1885,7 @@ static VALUE cJSON_parser_s_allocate(VALUE klass)
  */
 static VALUE cParser_source(VALUE self)
 {
-    GET_STRUCT;
+    GET_PARSER;
     return rb_str_dup(json->Vsource);
 }
 
@@ -1824,6 +1913,23 @@ void Init_parser()
     i_chr = rb_intern("chr");
     i_max_nesting = rb_intern("max_nesting");
     i_allow_nan = rb_intern("allow_nan");
+    i_symbolize_names = rb_intern("symbolize_names");
     i_object_class = rb_intern("object_class");
     i_array_class = rb_intern("array_class");
+    i_key_p = rb_intern("key?");
+    i_deep_const_get = rb_intern("deep_const_get");
+#ifdef HAVE_RUBY_ENCODING_H
+    CEncoding_UTF_8 = rb_funcall(rb_path2class("Encoding"), rb_intern("find"), 1, rb_str_new2("utf-8"));
+    CEncoding_UTF_16BE = rb_funcall(rb_path2class("Encoding"), rb_intern("find"), 1, rb_str_new2("utf-16be"));
+    CEncoding_UTF_16LE = rb_funcall(rb_path2class("Encoding"), rb_intern("find"), 1, rb_str_new2("utf-16le"));
+    CEncoding_UTF_32BE = rb_funcall(rb_path2class("Encoding"), rb_intern("find"), 1, rb_str_new2("utf-32be"));
+    CEncoding_UTF_32LE = rb_funcall(rb_path2class("Encoding"), rb_intern("find"), 1, rb_str_new2("utf-32le"));
+    CEncoding_ASCII_8BIT = rb_funcall(rb_path2class("Encoding"), rb_intern("find"), 1, rb_str_new2("ascii-8bit"));
+    i_encoding = rb_intern("encoding");
+    i_encode = rb_intern("encode");
+    i_encode_bang = rb_intern("encode!");
+    i_force_encoding = rb_intern("force_encoding");
+#else
+    i_iconv = rb_intern("iconv");
+#endif
 }
